
import Vec3D
import math
import sys

class CubicBezSurf:

    def __init__(self):
        self.Pts = []
        for i in xrange(10): self.Pts.append(Vec3D.Vec3D(0,0,0))

    def SurfFromGeomPts(self,pts):
        self._ControlPointsFromSurfPts(pts)

    def SurfFromGeomBdry(self,pts):
        self._ControlPointsFromBdry(pts)

    def SurfFromCntrlPts(self,pts):
        for i in xrange(10): self.Pts[i] = pts[i]

    def SurfFromNormals(self,corners,normals):
        self._ControlPointsFromNormals(corners,normals)

    def SurfFromMixedBdry(self,corners,normals,bdry,flags):
        self._ControlPointsFromMixedBdry(corners,bdry,flags,normals)

    def Normal(self,r,s=None):
        if s != None:
            dxdr,dxds = self.Derivatives(r,s)
        else:
            dxdr,dxds = self.Derivatives(r[0],r[1])
        return Vec3D.Vec3D(dxdr[1]*dxds[2] - dxds[1]*dxdr[2],
                           dxdr[2]*dxds[0] - dxds[2]*dxdr[0],
                           dxdr[0]*dxds[1] - dxds[0]*dxdr[1])

    def Reverse(self):
        tmp = [Pts[i] for i in xrange(10)]
        Pts[0] = tmp[9]
        Pts[1] = tmp[8]
        Pts[2] = tmp[5]
        Pts[3] = tmp[7]
        Pts[5] = tmp[2]
        Pts[7] = tmp[3]
        Pts[8] = tmp[1]
        Pts[9] = tmp[0]

    def Jacobian(self,r,s):
        dxdr,dxds = self.Derivatives(r,s)
        jacob = [[0,0,0],[0,0,0],[0,0,0]]
        for i in xrange(3):
            jacob[0][i] = dxdr[i]
            jacob[1][i] = dxds[i]
        jacob[2][0] = dxdr[1]*dxds[2] - dxds[1]*dxdr[2]
        jacob[2][1] = dxdr[2]*dxds[0] - dxds[2]*dxdr[0]
        jacob[2][1] = dxdr[0]*dxds[1] - dxds[0]*dxdr[1]
        return jacob

    def JacobDet(self,r,s):
        return self.Normal(r,s).Magnitude()

    def JacobInv(self,r,s):
        jac = self.Jacobian(r,s)
        inv = [[0,0,0],[0,0,0],[0,0,0]]
        det = math.sqrt(jac[2][0]*jac[2][0] +
                        jac[2][1]*jac[2][1] +
                        jac[2][2]*jac[2][2])
        inv[0][0] =  (jac[1][1]*jac[2][2] - jac[1][2]*jac[2][1]) / det
        inv[0][1] = -(jac[0][1]*jac[2][2] - jac[0][2]*jac[2][1]) / det
        inv[0][2] =  (jac[0][1]*jac[1][2] - jac[0][2]*jac[1][1]) / det
        inv[1][0] = -(jac[1][0]*jac[2][2] - jac[1][2]*jac[2][0]) / det
        inv[1][1] =  (jac[0][0]*jac[2][2] - jac[0][2]*jac[2][0]) / det
        inv[1][2] = -(jac[0][0]*jac[1][2] - jac[0][2]*jac[1][0]) / det 
        inv[2][0] =  (jac[1][0]*jac[2][1] - jac[1][1]*jac[2][0]) / det
        inv[2][1] = -(jac[0][0]*jac[2][1] - jac[0][1]*jac[2][0]) / det
        inv[2][2] =  (jac[0][0]*jac[1][1] - jac[0][1]*jac[1][0]) / det
        return inv

    def Evaluate(self,ri,si=None):
        if si == None:
            r = ri[0]
            s = ri[1]
        else:
            r = ri
            s = si
        t = 1 - r - s
        B = [0]*10
        B[0] = s*s*s
        B[1] = 3 * s*s * t
        B[2] = 3 * r * s*s
        B[3] = 3 * s * t*t
        B[4] = 6 * r * s * t
        B[5] = 3 * r*r * s
        B[6] = t*t*t
        B[7] = 3 * r * t*t
        B[8] = 3 * r*r * t
        B[9] = r*r*r
        x = Vec3D.Vec3D(0.0,0.0,0.0)
        for i in xrange(10): x += (self.Pts[i] * B[i])
        return x 

    def Derivatives(self,r,s):
        t = 1 - r - s
        dBdr = [0]*10
        dBdr[0] =  0
        dBdr[1] = -3 * s*s
        dBdr[2] =  3 * s*s
        dBdr[3] = -6 * s * t
        dBdr[4] =  6 * s * (t - r)
        dBdr[5] =  6 * r * s
        dBdr[6] = -3 * t*t
        dBdr[7] =  3 * t*t - 6 * r * t
        dBdr[8] =  6 * r * t - 3 * r*r
        dBdr[9] =  3 * r*r

        dBds = [0]*10
        dBds[0] =  3 * s*s
        dBds[1] =  6 * s * t - 3 * s*s
        dBds[2] =  6 * r * s
        dBds[3] =  3 * t*t - 6 * s * t
        dBds[4] =  6 * r * (t - s)
        dBds[5] =  3 * r*r
        dBds[6] = -3 * t*t
        dBds[7] = -6 * r * t
        dBds[8] = -3 * r*r
        dBds[9] =  0

        dxdr = Vec3D.Vec3D(0.0,0.0,0.0)
        dxds = Vec3D.Vec3D(0.0,0.0,0.0)
        for i in xrange(10):
            dxdr[0] += dBdr[i] * self.Pts[i][0] ;
            dxdr[1] += dBdr[i] * self.Pts[i][1] ;
            dxdr[2] += dBdr[i] * self.Pts[i][2] ;
            dxds[0] += dBds[i] * self.Pts[i][0] ;
            dxds[1] += dBds[i] * self.Pts[i][1] ;
            dxds[2] += dBds[i] * self.Pts[i][2] ;
        return dxdr,dxds

    def SecondDerivatives(self,r,s):
        t = 1 - r - s
        d2Bdr2 = [None]*10
        d2Bdrds = [None]*10
        d2Bds2 = [None]*10

        d2Bdr2[0] = 0
        d2Bdr2[1] = 0
        d2Bdr2[2] = 0
        d2Bdr2[3] = 6*s
        d2Bdr2[4] = -12*s
        d2Bdr2[5] = 6*s
        d2Bdr2[6] = 6*t
        d2Bdr2[7] = -12 + 18*r + 12*s
        d2Bdr2[8] = 6 - 18*r - 6*s
        d2Bdr2[9] = 6*r

        d2Bdrds[0] = 0
        d2Bdrds[1] = -6*s
        d2Bdrds[2] = 6*s
        d2Bdrds[3] = -6 + 6*r + 12*s
        d2Bdrds[4] = 6 - 12*r - 12*s
        d2Bdrds[5] = 6*r
        d2Bdrds[6] = 6*t
        d2Bdrds[7] = -6 + 12*r + 6*s
        d2Bdrds[8] = -6*r
        d2Bdrds[9] = 0

        d2Bds2[0] = 6*s
        d2Bds2[1] = 6 - 6*r - 18*s
        d2Bds2[2] = 6*r
        d2Bds2[3] = -12 + 12*r + 18*s
        d2Bds2[4] = -12*r
        d2Bds2[5] = 0
        d2Bds2[6] = 6*t
        d2Bds2[7] = 6*r
        d2Bds2[8] = 0
        d2Bds2[9] = 0

        d2xdr2  = Vec3D.Vec3D(0.0,0.0,0.0)
        d2xdrds = Vec3D.Vec3D(0.0,0.0,0.0)
        d2xds2  = Vec3D.Vec3D(0.0,0.0,0.0)

        for i in xrange(10):
            for j in xrange(3):
                d2xdr2[j]  += d2Bdr2[i]  * self.Pts[i][j]
                d2xdrds[j] += d2Bdrds[i] * self.Pts[i][j]
                d2xds2[j]  += d2Bds2[i]  * self.Pts[i][j]

        return d2xdr2,d2xdrds,d2xds2


    def FindClosestPoint(self,query,extrap=False,utol=-1):

        """
        This routine finds the point, X, on surface such that
        distance from X to query point, Q, is minimized.
        The algorithm employs a newton-raphson iteration on the
        following function (trying to find its zero):  

                 -             -   where  X(r,s) = element surface
        G(r,s) = | U dot dX/dr | ;        U(r,s) = X - Q
                 | U dot dX/ds |           (r,s) = spline local coords.
                 -             -  

        At a surface point at which U is normal to the surface, G=0.

        The newton step, S, is found from:  Jac(r,s) * S  =  -G(r,s)

        During the iteration, we make sure that we stay within the
        element, i.e., that (r,s) remains in range of [0.0, 1.0].

        The following heuristics are applied:
            (1) (Define a linear iteration step to be one in
                which the cross derivatives of X (d2X/drdr,
                d2X/drds, d2X/dsds) are set to zero.)
                If the off diagonal term of the jacobian is
                ever bigger than either of the diagonal
                terms, then drop down to a linear iteration.
                (This means that there is too much skewing occur-
                ring, so we are approximating the real element
                surface with one that has no skewing.)
            (2) Drop down into linear iteration mode for all
                steps after step 20.
            (3) Scale back all steps such that they remain
                less than (0.25).
            (4) Never take more than 30 steps.  Return the
                current location (and set status to FALSE)
                if the solution is not found after 30 steps.
        """

        MAX_STEPSIZE = 0.15   # max allow step size in any direc.
        MAX_STEPS = 30        # never take more than this many steps
        DROP_THRESHOLD = 20   # linear iteration after this many steps
        NEAR = 0.0001         # iteration stops when |cur-new|<NEAR

        if utol < 0.0:
            TOLER = 0.00001*(query[0]+query[1]+query[2])
        else:
                                                                                             
            # find the partial derivatives at the center of the
            # patch and the principal values in the jacobian.
 
            dXdr,dXds = Derivatives(0.3333333333,0.333333333)
            t0 = 0.5*(dXdr*dXdr + dXds*dXds)
            t1 = 0.5*(dXdr*dXdr - dXds*dXds)
            t2 = sqrt(t1*t1 + dXdr*dXds)
            dXm0 = t0 + t2
            dXm1 = t0 - t2

            if dXm0 < dXm1:
                TOLER = dXm0 * utol
            else:
                TOLER = dXm1 * utol
            NEAR = utol
 
        # Start the iteration at 0,0,0

        newp = Vec3D.Vec3D(0.0,0.0,0.0)
        stepnum = 0
        linear_mode = False

        done = False
        while not done:
            stepnum += 1
            cur = newp

            # If we've taken too many steps, then break out of loop

            if stepnum > MAX_STEPS:
                dist = (closest-query).Magnitude()
                r = newp[0]
                s = newp[1]
                return False,closest,r,s,dist

            # Drop down into linear iteration mode if too many steps have
            # been taken.

            if stepnum > DROP_THRESHOLD: linear_mode = True

            # Compute element geometry and local derivatives at
            # current location.

            closest = self.Evaluate(cur[0],cur[1])
            dXdr,dXds = self.Derivatives(cur[0],cur[1])
            d2Xdr2,d2Xdrds,d2Xds2 = self.SecondDerivatives(cur[0],cur[1])

            # Compute U function at current location

            U = closest - query

            # if the closest and query point are the same,
            # return the current point

            if abs(U[0]) <= TOLER and abs(U[1]) <= TOLER and abs(U[2]) <= TOLER:
                dist = U.Magnitude()
                r = newp[0]
                s = newp[1]
                return True,closest,r,s,dist

            # Evaluate G at current location

            G = [None]*2
            G[0] = U*dXdr
            G[1] = U*dXds

            # Evaluate the jacobian at current location.
            # (If linear_mode flag is set, do it in linear mode; otherwise,
            # only do it in linear mode if the off-diagonal term is too big.)

            jac = [[dXdr*dXdr,dXdr*dXds],[dXdr*dXds,dXds*dXds]]

            if not linear_mode:
                jac[0][0] += d2Xdr2*U
                jac[0][1] += d2Xdrds*U
                jac[1][0] = jac[0][1]
                jac[1][1] += d2Xds2*U
                if abs(jac[0][1]) > abs(jac[0][0]) or abs(jac[0][1]) > abs(jac[1][1]):
                    jac[0][0] = dXdr*dXdr
                    jac[0][1] = dXdr*dXds
                    jac[1][0] = jac[0][1]
                    jac[1][1] = dXds*dXds

            # Invert the jacobian.

            det_jac = jac[0][0]*jac[1][1] - jac[0][1]*jac[1][0]
            inv_jac = [[jac[1][1]/det_jac,-jac[0][1]/det_jac],
                       [-jac[1][0]/det_jac,jac[0][0]/det_jac]]

            # Compute the step.

            step = Vec3D.Vec3D(
                    -(inv_jac[0][0] * G[0] + inv_jac[0][1] * G[1]), 
                    -(inv_jac[1][0] * G[0] + inv_jac[1][1] * G[1]),0.0)

            # Scale back this step in case it is "too big".

            if step.Magnitude() > MAX_STEPSIZE:
                scalefac = MAX_STEPSIZE / step.Magnitude()
                step *= scalefac

            newp = cur + step

            # Pull the new location back (if necessary) to insure that it
            # lies inside the element.

            if not extrap:
                if newp[0] > 1.0: newp[0] = 1.0
                if newp[0] < 0.0: newp[0] = 0.0
                if newp[1] > 1.0: newp[1] = 1.0
                if newp[1] < 0.0: newp[1] = 0.0

            # check for convergance

            if abs(newp[0] - cur[0]) <= NEAR and abs(newp[1] - cur[1]) <= NEAR:
                done = True ;

        return True,closest,newp[0],newp[1],(closest-query).Magnitude()


    def _ControlPointsFromNormals(self,corners,norms):

        # first the corner points

        self.Pts[0] = corners[0]
        self.Pts[6] = corners[1]
        self.Pts[9] = corners[2]

        # determine the outside control points

        self.Pts[1],self.Pts[3] = self._FindMidCntrl(corners[0],corners[1],
                                                     norms[0],norms[1])
        self.Pts[7],self.Pts[8] = self._FindMidCntrl(corners[1],corners[2],
                                                     norms[1],norms[2])
        self.Pts[5],self.Pts[2] = self._FindMidCntrl(corners[2],corners[0],
                                                     norms[2],norms[0])

        # and the middle point

        self.Pts[4] = ((self.Pts[1]+self.Pts[3]+self.Pts[7]+
                        self.Pts[8]+self.Pts[5]+self.Pts[2])/4.0 -\
                       (self.Pts[0]+self.Pts[6]+self.Pts[9])/6.0)


#       input point ordering for       point ordering for
#       boundary geom points           all geom points
#
#               0 o                              0 o
#                / \                              / \
#               /   \                            /   \
#            3 o-----o 8                      1 o-----o 2
#             / \   / \                        / \   / \
#            /   \ /   \                      /   \ /   \
#         4 o-----o-----o 7                3 o-----4-----o 5
#          / \   / \   / \                  / \   / \   / \
#         /   \ /   \ /   \                /   \ /   \ /   \
#      1 o-----o-----o-----o 2          6 o-----o-----o-----o 9
#              5     6                          7     9
#

    def _ControlPointsFromBdry(self,bdry):

        lmap = [[1,3],[7,8],[5,2]]

        # first the corner points

        self.Pts[0] = bdry[0]
        self.Pts[6] = bdry[1]
        self.Pts[9] = bdry[2]

        # use the side point locations to determine the bdry control points

        for i in xrange(3):
            j = 2*i + 3
            k = (i+1) % 3
            self.Pts[lmap[i][0]] = -bdry[i]/1.2 + 3.0*bdry[j] - \
                                   1.5*bdry[j+1] + bdry[k]/3.0
            self.Pts[lmap[i][1]] =  bdry[i]/3.0 - 1.5*bdry[j] + \
                                   3.0*bdry[j+1] - bdry[k]/1.2

        self.Pts[4] = ((self.Pts[1]+self.Pts[3]+self.Pts[7]+
                        self.Pts[8]+self.Pts[5]+self.Pts[2])/4.0 - \
                       (self.Pts[0]+self.Pts[6]+self.Pts[9])/6.0) ;


    def _ControlPointsFromMixedBdry(self,corners,bdry,bdry_flags,norms):

        lmap = ((1,3),(7,8),(5,2))

        # first the corner points

        self.Pts[0] = corners[0]
        self.Pts[6] = corners[1]
        self.Pts[9] = corners[2]

        # use the side point locations to determine the bdry
        # control points

        for i in xrange(3):
            if bdry_flags[i]:
                k = (i+1) % 3
                self.Pts[lmap[i][0]] = -corners[i]/1.2 + 3.0*bdry[i][0] - \
                                       1.5*bdry[i][1] + corners[k]/3.0
                self.Pts[lmap[i][1]] =  corners[i]/3.0 - 1.5*bdry[i][0] + \
                                       3.0*bdry[i][1] - corners[k]/1.2
            else:
                j = (i+1) % 3
                self.Pts[lmap[i][0]],self.Pts[lmap[i][1]] = \
                    self._FindMidCntrl(corners[i],corners[j],norms[i],norms[j])

        self.Pts[4] = ((self.Pts[1]+self.Pts[3]+self.Pts[7]+
                        self.Pts[8]+self.Pts[5]+self.Pts[2])/4.0 -
                       (self.Pts[0]+self.Pts[6]+self.Pts[9])/6.0)


    def _ControlPointsFromSurfPts(self,spts):

        ZERO =  0.0
        ONE  =  1.0
        THRE =  3.0
        FCT1 = -0.833333333333333333
        FCT2 =  0.333333333333333333
        FCT3 = -1.5
        FCT4 = -0.75
        FCT5 =  4.5

        coef = [
            [  ONE, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO ],
            [ FCT1, THRE, ZERO, FCT3, ZERO, ZERO, FCT2, ZERO, ZERO, ZERO ],
            [ FCT1, ZERO, THRE, ZERO, ZERO, FCT3, ZERO, ZERO, ZERO, FCT2 ],
            [ FCT2, FCT3, ZERO, THRE, ZERO, ZERO, FCT1, ZERO, ZERO, ZERO ],
            [ FCT2, FCT4, FCT4, FCT4, FCT5, FCT4, FCT2, FCT4, FCT4, FCT2 ],
            [ FCT2, ZERO, FCT3, ZERO, ZERO, THRE, ZERO, ZERO, ZERO, FCT1 ],
            [ ZERO, ZERO, ZERO, ZERO, ZERO, ZERO,  ONE, ZERO, ZERO, ZERO ],
            [ ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, FCT1, THRE, FCT3, FCT2 ],
            [ ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, FCT2, FCT3, THRE, FCT1 ],
            [ ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO, ZERO,  ONE ]]

        for i in xrange(10):
            self.Pts[i] = Vec3D.Vec3D(0,0,0)
            for j in xrange(10):
                self.Pts[i] += coef[i][j] * spts[j]


    def _FindMidCntrl(self,corner0,corner1,norm0,norm1):

        # Cross the end points with the plane normal to get a vector in
        # the normal plane.  Then cross again to get the projection
        # of the sector vector onto the normal plane

        sec_vec = corner1 - corner0 ;
        tmp = Vec3D.CrossProd(sec_vec,norm0)
        dn0 = Vec3D.CrossProd(norm0,tmp).Normalize()

        sec_vec = -sec_vec
        tmp = Vec3D.CrossProd(sec_vec,norm1)
        dn1 = Vec3D.CrossProd(norm1,tmp).Normalize()

        # Now compute the coefficients for the quadratic
        # equation and solve for the control point distances

        at = dn0 - dn1
        a = (at*at) - 1.0
        b = 2*corner0*(dn0-dn1) + 2*corner1*(dn1-dn0)
        ct = corner0-corner1
        c = ct*ct

        t = (-b - math.sqrt(b*b - 4.0*a*c))/(2.0*a)

        # Now the control points

        cp1 = corner0 + t * dn0
        cp2 = corner1 + t * dn1

        return cp1,cp2


    def DebugDrawSurface(self,fd,num=5):

        pts = []
        for i in xrange(num):
            tmp = []
            for i in xrange(num): tmp.append(Vec3D.Vec3D(0,0,0))
            pts.append(tmp)


        for i in xrange(num):
            for j in xrange(num-i):
                pts[i][j] = self.Evaluate(float(j/(num-1.0)),
                                          float(i/(num-1.0)))

        for i in xrange(num-1):
            for j in xrange(num-i-1):
                self._DebugDrawPoly(pts[i][j],pts[i][j+1],pts[i+1][j],fd)

        for i in xrange(num-1):
            for j in xrange(1,num-i-1):
                self._DebugDrawPoly(pts[i][j],pts[i+1][j],pts[i+1][j-1],fd)



    def DebugDrawOutline(self,fd,num=5):
        for i in xrange(num):
            x0 = float(i) / float(num) ;
            x1 = float(i+1) / float(num) ;
            print >> fd,"e",self.Evaluate(x0,0),self.Evaluate(x1,0)
            print >> fd,"e",self.Evaluate(0,x0),self.Evaluate(0,x1)
            print >> fd,"e",self.Evaluate(1-x0,x0),self.Evaluate(1-x1,x1)


    def _DebugDrawPoly(self,pt0,pt1,pt2,fd):
        print >> fd,"p 3",pt0,pt1, pt2
        #print >> sys.stderr,pt0,pt1, pt2
