#!/usr/bin/python

import sys
import math
import string

import Vec3D
import CubicBezSurf
import FlawSurfs


NONE = 0
ANSYS = 1
ABAQUS = 2


nodes = {}
quad_elems = {}
tri_elems = {}
comps = {}


def GetAnsysLine(fd):
    """this is a little helper function that scans the input
       file and ignores lines starting with a '/'"""
    buff = fd.readline()
    if len(buff) == 0: return None
    while buff[0] == '/' or buff[0] == '*' :
        buff = fd.readline()
        if len(buff) == 0: return None
    return buff


def IsInt( str ):
    """ Is the given string an integer? """
    ok = 1
    try:
        num = int(str)
    except ValueError:
        ok = 0
    return ok
    

def ReadAnsys(fd):
    """this is a function that reads ANSYS cdb files"""

    fd.seek(0)
    buff = GetAnsysLine(fd)
    while buff != None:
        vals = string.split(buff,",")

        if vals[0].lower() == 'nblock' :
            # read the node coordinates
            buff = GetAnsysLine(fd) # format line
            buff = GetAnsysLine(fd)
        
            while buff != None:
        
            tmp = ""
        pc = ""
            for ic in buff:
                    if ( ic=='\r' or ic=='\n' or ic=='\0' ): break
                    if ( ic == '-' and IsInt(pc) ): tmp = tmp + ' '
                    tmp = tmp + ic
            pc = ic         
                tmp = tmp + '\n' ;
            buff = tmp
        
            vals = string.split(buff)
            if vals[1] == '-1' or vals[1] == '-1,' :
                break
            else:
                if len(vals) >= 6:
                nodes[int(vals[0])] = (float(vals[3]),float(vals[4]),float(vals[5]))
                elif len(vals) == 5:
                nodes[int(vals[0])] = (float(vals[3]),float(vals[4]),float(0.0))
                elif len(vals) == 4:
                nodes[int(vals[0])] = (float(vals[3]),float(0.0),float(0.0))
            buff = GetAnsysLine(fd)
            #end while NBLOCK
            
        elif vals[0].lower() == 'eblock' :
            # read the element definition
            buff = GetAnsysLine(fd) # format line
            buff = GetAnsysLine(fd)
            while buff != None:
            vals = string.split(buff)
            if vals[0] == '-1' :
                break
            else:
                if int(vals[8]) == 4 and vals[13] != vals[14]:
                quad_elems[vals[10]] = (int(vals[11]),int(vals[12]),int(vals[13]),int(vals[14]))
                else:
                tri_elems[vals[10]] = (int(vals[11]),int(vals[12]),int(vals[13]))
            buff = GetAnsysLine(fd)
            #end while EBLOCK

        elif vals[0].lower() == 'cmblock' :
        num = int(string.split(vals[3])[0])
        name = vals[1]
            # read the component definition
            buff = GetAnsysLine(fd) # format line
            buff = GetAnsysLine(fd)
        cnodes = []
            while buff != None:
            vals = string.split(buff)
        for v in vals:
            cnodes.append(v)
        if len(cnodes) == num :
                    break
            buff = GetAnsysLine(fd)
            #end while CMBLOCK   
            comps[name] = cnodes 


        buff = GetAnsysLine(fd)
    
        #end while read cdb file 

    return


def GetAbaqusLine(fd):
    """this is a little helper function that scans the input
       file and ignores lines starting with a '**'"""
    buff = fd.readline()
    if len(buff) == 0: return None
    while buff[0] == '*' and buff[1] == '*' :
        buff = fd.readline()
        if len(buff) == 0: return None
    return buff


def ReadAbaqus(fd):
    """this is a function that reads ABAQUS inp files"""
    
    fd.seek(0)
    buff = GetAbaqusLine(fd)
    while buff != None:
        read_next = 1
        tmp = string.replace(buff,","," ")
        vals = string.split(tmp)

        if vals[0].lower()[:5] == '*node' :
            # read the node coordinates
            buff = GetAbaqusLine(fd)
        
            while buff != None:
        
                tmp = string.replace(buff,","," ")
                vals = string.split(tmp)

            if buff == None or buff[0] == '*' :
            #print "break from node block"
            read_next = 0
                break
            else:
            #print vals
                if len(vals) == 4:
                nodes[int(vals[0])] = (float(vals[1]),float(vals[2]),float(vals[3]))
                elif len(vals) == 3:
                nodes[int(vals[0])] = (float(vals[1]),float(vals[2]),float(0.0))
                elif len(vals) == 2:
                nodes[int(vals[0])] = (float(vals[1]),float(0.0),float(0.0))
            buff = GetAbaqusLine(fd)
            #end while *Node
            
        elif vals[0].lower()[:8] == '*element' :
        if len(vals) > 1:
            if vals[1].lower()[:4] != "type":
            buff = GetAbaqusLine(fd)
            continue 
        
            # read the element definition
            buff = GetAbaqusLine(fd)
        
        while buff != None:
        
                tmp = string.replace(buff,","," ")
                vals = string.split(tmp)

            if buff == None or buff[0] == '*' :
            #print "break from elem block"
            read_next = 0
                break
            else:
                if len(vals) == 9 or len(vals) == 5 :
                quad_elems[vals[0]] = (int(vals[1]),int(vals[2]),int(vals[3]),int(vals[4]))
                else:
                tri_elems[vals[0]] = (int(vals[1]),int(vals[2]),int(vals[3]))
            buff = GetAbaqusLine(fd)
            #end while *Element

        elif vals[0].lower()[:5] == '*nset' :
        generate_nodes = 0
        name = vals[1][5:].lower()
        if len(vals) == 3:
            if vals[2].lower() == "generate":
            generate_nodes = 1
        
            # read the component definition
            buff = GetAbaqusLine(fd)
        cnodes = []
            while buff != None:
        
                tmp = string.replace(buff,","," ")
                vals = string.split(tmp)

            if buff == None or buff[0] == '*' :
            #print "break from nset block"
            read_next = 0
                break
            else:
            if generate_nodes ==0:
                for v in vals:
                    cnodes.append(v)
                else:
                start = int(vals[0])
            end = int(vals[1])
            increment = int(vals[2])
            for i in range(start,end,increment):
                cnodes.append(i)
            cnodes.append(end)

            buff = GetAbaqusLine(fd)
            #end while *Nset
            comps[name] = cnodes


        if ( read_next == 1 ): buff = GetAbaqusLine(fd)
    
        #end while read inp file 

    return



if __name__ == '__main__':

    if len(sys.argv) < 3:
        print >> sys.stderr,"usage: python FESurfCrack.py -tr 0.005 -cdb name.cdb > f3d.crk"
        print >> sys.stderr,"usage: python FESurfCrack.py -tr 0.005 -inp name.inp > f3d.crk"
    sys.exit(0)
    
    verbose = None
    if "-v" in sys.argv: verbose = 1

    DefRad = 0.01
    fname = None
    fem_code = NONE
    
    for i in xrange(len(sys.argv)):
        if sys.argv[i] == "-tr":
        if i < len(sys.argv):
                DefRad = float(sys.argv[i+1])
        if sys.argv[i] == "-cdb":
        if i < len(sys.argv):
                fname = sys.argv[i+1]
        fem_code = ANSYS
        if sys.argv[i] == "-inp":
        if i < len(sys.argv):
                fname = sys.argv[i+1]
        fem_code = ABAQUS

    # read the input file
    if fname != None:
        fd = open(fname, 'r')
    else:
        print >> sys.stderr,"no input file specified"
        sys.exit(0)
  
    if fem_code == ANSYS:
        print >> sys.stderr,"read ansys input file"
        ReadAnsys(fd)
    elif fem_code == ABAQUS:
        print >> sys.stderr,"read abaqus input file"
        ReadAbaqus(fd)
    else:
        print >> sys.stderr,"invalid input file"
        sys.exit(0) 

    print >> sys.stderr,"number of nodes: ", len(nodes)
    print >> sys.stderr,"number of quad elements: ", len(quad_elems)
    print >> sys.stderr,"number of tri elements: ", len(tri_elems)
    print >> sys.stderr,"number of components: ", len(comps)
    
    # store the element edges that make up the mesh boundary
    # if an edge is used twice, it is not part of the boundary
    # and thus it is removed from the dictionary

    edges = {}
    
    quad_edges = [(0,1),(1,2),(2,3),(3,0)]
    tri_edges = [(0,1),(1,2),(2,0)]

    for el in quad_elems:
        elem = quad_elems[el]
    for i in xrange(len(quad_edges)):
        edge = []
        for j in xrange(2):
               edge.append(elem[quad_edges[i][j]])

            edge.sort()
            edge = (edge[0],edge[1])
        if edge in edges:
            del edges[edge]
            else:
            edges[edge] = el

    for el in tri_elems:
        elem = tri_elems[el]
    for i in xrange(len(tri_edges)):
        edge = []
        for j in xrange(2):
               edge.append(elem[tri_edges[i][j]])

            edge.sort()
            edge = (edge[0],edge[1])
        
        if edge in edges:
            del edges[edge]
            else:
            edges[edge] = el
    

    # create ordered set of edges and nodes along the boundaries
    # there can be multiple boundaries and closed loop boundaries
    
    ordered_nodes = []
    ordered_edges = []
    
    edge_list = edges.keys()
        
    while len(edge_list) > 0:

        # collect the edges on a boundary 
    # 'randomly' start from first node in list
    
        ord_nodes = []
        ord_edges = []

        start_node = edge_list[0][0]
        next_node = edge_list[0][1]

        # start a boundary

        ord_nodes.append(start_node)
        ord_nodes.append(next_node)
        ord_edges.append(edge_list[0])
    
        while next_node != start_node:
            for edge in edge_list:
            if edge != ord_edges[-1]:
                if next_node == edge[0]:
                ord_nodes.append(edge[1])
                    next_node = edge[1]
                ord_edges.append(edge)
            elif next_node == edge[1]:
                ord_nodes.append(edge[0])
                    next_node = edge[0]
                ord_edges.append(edge)
        # end while next_node
    
    # save the boundary in the list
    
        ordered_nodes.append(ord_nodes)
        ordered_edges.append(ord_edges)
    
    # loop through the edges and see if all are used
    # if any are not used in the current boundary
    # make a new list so we can collect the next boundary
    
    new_edge_list = []
        for edge in edge_list:
            if edge not in ord_edges:
                new_edge_list.append(edge)
        edge_list = new_edge_list

    # end while len(edge_list) -> all boundaries found


    # if there are components, collect all node ids for each
    
    front_nodes = []
    # front_nodes is a list of lists of nodes on all fronts

    for c in comps:
    cmn = comps[c]
        front = []
    for i in xrange(len(cmn)-1):
        start = int(cmn[i])
        end = int(cmn[i+1])
        # component listings use a range indicated by -sign
        if end < 0 :
            stop = -end
            for n in xrange(stop-start):
            front.append(start+n)
            front.append(stop)
        i = i+1
        else :
            if start > 0 :
                front.append(start)
                if ( i+1 == len(cmn)-1 ):
                    front.append(end)
    front_nodes.append(front)

    # end of check for components


    # verify that all nodes in the front are in the 
    # set of ordered boundary nodes (removes midsides)

    tmp_fronts = []
    for front in front_nodes:
        tmp = []
        for node in front:
        for bound in ordered_nodes:
            if node in bound:
            tmp.append(node)
            
        tmp_fronts.append(tmp)

    front_nodes = tmp_fronts


    # do some checks on the fronts
    
    tmp_fronts = []
    for front in front_nodes:
        next_front = 0
        for node in front:
    
        # look for the front node in the ordered boundaries
        for bound in ordered_nodes:
        
            if node in bound:
        
                    # front node found in current boundary
            if len(front) == len(bound):
                print >> sys.stderr, "front is assumed to be full boundary"
            tmp_fronts.append(bound)
            elif len(front) == len(bound)-1 and bound[0] == bound[len(bound)-1]:
                print >> sys.stderr, "front is assumed to be full closed boundary"
            tmp_fronts.append(bound)
            else:
                print >> sys.stderr, "front is part of boundary; you must ensure ordering"
            tmp_fronts.append(front)

            # front node found in boundary - check next front 
            next_front = 1
                    break

            if next_front == 1: break


    front_nodes = tmp_fronts


    # define the triangular flaw surfaces
 
    corners = [None]*3
    normals = [None]*3
    quad_tri = [(0,1,2),(0,2,3)]
    
    surfs = []
    corner_nds = []
    
    # break the quad elements into two triangles
        
    for el in quad_elems:
        elem = quad_elems[el]
    for i in xrange(len(quad_tri)):
        nd_ids = []
        for j in xrange(3):
            nd_ids.append(elem[quad_tri[i][j]])
            nd = nodes[elem[quad_tri[i][j]]]
                crd = Vec3D.Vec3D(nd[0],nd[1],nd[2])
                corners[j] = crd
        
            corner_nds.append(nd_ids)
            v0 = (corners[1] - corners[0]).Normalize()
            v1 = (corners[2] - corners[0]).Normalize()
            norm = Vec3D.CrossProd(v0,v1).Normalize()
            for j in xrange(3):
            normals[j] = norm
        
            surf = CubicBezSurf.CubicBezSurf()
            surf.SurfFromNormals(corners,normals)
            surfs.append(surf)

    for el in tri_elems:
        elem = tri_elems[el]
    nd_ids = []
    for j in xrange(3):
        nd_ids.append(elem[j])
        nd = nodes[elem[j]]
            crd = Vec3D.Vec3D(nd[0],nd[1],nd[2])
            corners[j] = crd

        corner_nds.append(nd_ids)
        v0 = (corners[1]-corners[0]).Normalize()
        v1 = (corners[2]-corners[0]).Normalize()
        norm = Vec3D.CrossProd(v0,v1).Normalize()
        for j in xrange(3):
        normals[j] = norm
        
        surf = CubicBezSurf.CubicBezSurf()
        surf.SurfFromNormals(corners,normals)
        surfs.append(surf)


    # add the triangles into the FlawSurf object 
    
    fs = FlawSurfs.FlawSurfs()
    for i in xrange(len(surfs)):
        fs.AddSurf(surfs[i],corner_nds[i][0],corner_nds[i][1],corner_nds[i][2])
    fs.MakeWatertight()

    # compare the front node order with the flaw surfs
    # reverse front node order if needed
    
    tmp_fronts = []
    for front in front_nodes:
        for cn in corner_nds:
            if cn[0] in front and cn[1] in front:
        ind1 = front.index(cn[0])
        ind2 = front.index(cn[1])
        elif cn[1] in front and cn[2] in front:
        ind1 = front.index(cn[1])
        ind2 = front.index(cn[2])
        elif cn[2] in front and cn[0] in front:
        ind1 = front.index(cn[2])
        ind2 = front.index(cn[0])
        if ind1 > ind2:
            front.reverse()

    tmp_fronts.append(front)

    front_nodes = tmp_fronts


    # define the crack front and template parameters
    # only process one crack front for now
    for front in front_nodes:
        tmp = []
        if len(front) > 0:
            for i in xrange(len(front)): 
                tmp.append(DefRad)
            fs.AddFront(front,tmp,fs.TempParams())

    # write the flaw to stdout: usually piped to file
    
    fs.Save(sys.stdout)
