
import sys
import math
import Vec3D
import Vec2D
import CubicBezSurf
import BSpline3D

class FileIter:

    def __init__(self,fd):
        self.fd = fd
        self.v = []

    def __getitem__(self,i):
        return self.v[i]

    def __len__(self):
        return len(self.v)

    def Next(self):
        self.v = []
        buff = self.fd.readline()
        if len(buff) > 0: self.v = buff.split()

    def AsInt(self,i):
        return int(self.v[i])

    def AsFloat(self,i):
        return float(self.v[i])

def EdgeKey(id0,id1):
    if id0 < id1: return (id0,id1)
    return (id1,id0)

#class EdgeKey:
#
#    def __init__(self,id0,id1):
#        if id0 < id1:
#            self.id = (id0,id1)
#        else:
#            self.id = (id1,id0)
#
#    def __hash__(self):
#        print >> sys.stderr,"in hash"
#        return self.id.__hash__()
#
#    def __repr__(self):
#        #return "(%d,%d)" % (self.id[0],self.id[1]) 
#        return "%d" % self.id.__hash__()

#
#   FlawSurfs.py - python implementation of the FlawSurfs object
#
#
#   The assumed order of the corner nodes and control points is:
#
#                o 0  <- corner 0
#                |\
#                | \
#              1 o  o 2
#                |   \
#                |  4 \
#              3 o  o  o 5
#                |      \
#                |       \
#    corner 1 -> o--o--o--o <- corner 2
#               6   7  8   9
#


class FlawSurfs:

    # embedded classes

    class Edge:

        def __init__(self,ivt0,ivt1,iface0,iface1):
            self.vt0 = ivt0
            self.vt1 = ivt1
            self.face0 = iface0
            self.face1 = iface1
            self.side_node = -1
            self.face0_par = [None]*2
            self.face1_par = [None]*2

    class EdgePts:

        def __init__(self,pt0,pt1,face):
            self.pt0 = pt0
            self.pt1 = pt1
            self.face = face


    class TempParams:

        def __init__(self,progression_ratio=1,num_rings=3,num_circ=8,
                     max_aspect_ratio=2,transition_factor=4,
                     do_transitions=False,do_simple=False):
            self.ProgressionRatio = progression_ratio
            self.NumRings = num_rings
            self.NumCirc = num_circ
            self.MaxAspectRatio = max_aspect_ratio
            self.NormalTransitionFactor = transition_factor
            self.DoTransitions = do_transitions
            self.SimpleIntersections = do_simple


    class VtxRef:

        def __init__(self,face,param):
            self.face = face
            self.param = param


    # methods

    def __init__(self):
        self.Dirty = False
        self.Fronts = []
        self.FrontTempRadii = []
        self.FrontTempParams = []
        self.Segments = []
        self.Nodes = {}
        self.Surfs = {}
        self.NextNodeId = 0
        self.NextSurfId = 0
        self.NodeCds = {}
        self.NodeNrm = {}
        self.EdgeMap = (
            ((0.0,1.0),(0.0,0.666667),(0.0,0.333333),(0.0,0.0)),
            ((0.0,0.0),(0.333333,0.0),(0.666667,0.0),(1.0,0.0)),
            ((1.0,0.0),(0.666667,0.333333),(0.333333,0.666667),(0.0,1.0)))

    def AddSurf(self,surf,ind0=None,ind1=None,ind2=None):
        tol = (surf.Pts[6]-surf.Pts[0]).Magnitude()
        tol += (surf.Pts[9]-surf.Pts[6]).Magnitude()
        tol += (surf.Pts[0]-surf.Pts[9]).Magnitude()
        tol = 0.001 * tol/3.0

        if ind0 != None:
            nd0 = ind0
        else:
            nd0 = self.FindVtxId(surf.Pts[0],tol)
            if nd0 == None:
                nd0 = self.NextNodeId
                self.NextNodeId += 1
                self.NodeCds[nd0] = surf.Pts[0]
                self.NodeNrm[nd0] = surf.Normal(0,1)
            else:
                surf.Pts[0] = self.NodeCds[nd0]

        if ind1 != None:
            nd1 = ind1
        else:
            nd1 = self.FindVtxId(surf.Pts[6],tol)
            if nd1 == None:
                nd1 = self.NextNodeId
                self.NextNodeId += 1
                self.NodeCds[nd1] = surf.Pts[6]
                self.NodeNrm[nd1] = surf.Normal(0,0)
            else:
                surf.Pts[6] = self.NodeCds[nd1]

        if ind2 != None:
            nd2 = ind2
        else:
            nd2 = self.FindVtxId(surf.Pts[9],tol)
            if nd2 == None:
                nd2 = self.NextNodeId
                self.NextNodeId += 1
                self.NodeCds[nd2] = surf.Pts[9]
                self.NodeNrm[nd2] = surf.Normal(1,0)
            else:
                surf.Pts[9] = self.NodeCds[nd2]

        #print >> sys.stderr,self.NextSurfId,nd0,nd1,nd2

        self.Nodes[self.NextSurfId] = (nd0,nd1,nd2)
        self.Surfs[self.NextSurfId] = surf
        self.NextSurfId += 1
        if nd0 >= self.NextNodeId: self.NextNodeId = nd0 + 1
        if nd1 >= self.NextNodeId: self.NextNodeId = nd1 + 1
        if nd2 >= self.NextNodeId: self.NextNodeId = nd2 + 1
        self.Dirty = True

    def AddFront(self,front,temp_radii,temp_params):
        self.Fronts.append(front)
        self.FrontTempRadii.append(temp_radii)
        self.FrontTempParams.append(temp_params)
        self.Dirty = True ;

    def AddFrontUnifRadius(self,front,temp_radius,temp_params=None):
        if temp_params == None: temp_params = self.TempParams()
        self.Fronts.append(front)
        tmp = [temp_radius for i in xrange(len(front))]
        self.FrontTempRadii.append(tmp)
        self.FrontTempParams.append(temp_params)
        self.Dirty = True

    def RemoveSurf(self,sid):
        del self.Surfs[sid]
        del self.Nodes[sid]

    def RemoveFronts(self):
        self.Fronts = []
        self.FrontTempRadii = []
        self.FrontTempParams = []

    def Clear(self):
        self.Fronts = []
        self.FrontTempRadii = []
        self.FrontTempParams = []
        self.Nodes = {}
        self.Surfs = {}
        self.NodeCds = {}
        self.NodeNrm = {}
        self.Segments = []
        self.NextNodeId = 0
        self.Dirty = True

    def HaveFlawSurfs(self):
        return self.Surfs.has_key(Surfs)

    def HaveFlawFronts(self):
        return len(self.Fronts) > 0

    def NodeCoord(self,i):
        #if len(self.NodeCds) == 0: self._GenerateNodeCds()
        tmp = self.NodeCds[i]
        return Vec3D.Vec3D(tmp[0],tmp[1],tmp[2])

    def NodeNormal(self,i):
        #if len(self.NodeNrm) == 0: self._GenerateNodeCds()
        tmp = self.NodeCds[i]
        return Vec3D.Vec3D(tmp[0],tmp[1],tmp[2])

    def FindVtxId(self,coord,tol):
        #if len(self.NodeNrm) == 0: self._GenerateNodeCds()
        for nid in self.NodeCds:
            if (self.NodeCds[nid]-coord).Magnitude() < tol: return nid
        return None

    def SetFrontTempParams(self,fid,template_params):
        self.FrontTempParams[fid] = template_params
        self.Dirty = true

    def Save(self,fd):
        self.MakeWatertight()
        print >> fd, "FLAWSURF"
        print >> fd, "("
        print >> fd, "VERSION: ", 1
        print >> fd, "NUM_SURFS: ", len(self.Nodes)

        for i in self.Nodes:
            nd = self.Nodes[i]
            print >> fd, "SURF: ", i
            print >> fd, "    ", nd[0], nd[1], nd[2]

            pts = self.Surfs[i].Pts

            for j in xrange(10):
                print >> fd, "    %22.15g %22.15g %22.15g" % (pts[j][0],pts[j][1],pts[j][2])

        print >> fd, "NUM_FRONTS: ", len(self.Fronts)

        for i in xrange(len(self.Fronts)):
            print >> fd, "FRONT: ", i, len(self.Fronts[i]),
            for j in xrange(len(self.Fronts[i])):
                if (j%10) == 0: print >> fd
                print >> fd, self.Fronts[i][j],
            print >> fd

            print >> fd, "RADII: ", i, len(self.Fronts[i]),
            for j in xrange(len(self.Fronts[i])):
                if (j%10) == 0: print >> fd
                print >> fd, self.FrontTempRadii[i][j],
            print >> fd

            if self.FrontTempParams[i].DoTransitions:
                tmp = 1
            else:
                tmp = 0

            print >> fd, "TEMP_PAR: ", i, 
            print >> fd, self.FrontTempParams[i].ProgressionRatio,
            print >> fd, self.FrontTempParams[i].NumRings,
            print >> fd, self.FrontTempParams[i].NumCirc,
            print >> fd, self.FrontTempParams[i].MaxAspectRatio,
            print >> fd, self.FrontTempParams[i].NormalTransitionFactor,
            print >> fd, tmp

        print >> fd, ")"

    def Load(self,fd):
        fr = FileIter(fd)
        self.Fronts = []
        self.Nodes = {}
        self.Surfs = {}

        fr.Next()
        if fr[0] == "FLAWSURF": fr.Next()
        fr.Next()   # read the opening "("

        if fr[0] != "VERSION:": raise "FormatException"
        version = fr.AsInt(1)
        if version < 1 or version > 2:    raise "VersionException"

        fr.Next()
        if fr[0] != "NUM_SURFS:": raise "FormatException"
        num_surfs = fr.AsInt(1)

        for i in xrange(num_surfs):
            fr.Next()
            if fr[0] != "SURF:": raise "FormatException"
            nid = fr.AsInt(1)
            if nid >= self.NextSurfId: self.NextSurfId = nid+1
            fr.Next()
            self.Nodes[nid] = (fr.AsInt(0),fr.AsInt(1),fr.AsInt(2))
            for j in xrange(3):
                if self.Nodes[nid][j] >= self.NextNodeId:
                    self.NextNodeId = self.Nodes[nid][j]+1

            pts = []
            for j in xrange(10):
                fr.Next()
                pts.append(Vec3D.Vec3D(fr.AsFloat(0),fr.AsFloat(1),fr.AsFloat(2)))

            self.Surfs[nid] = CubicBezSurf.CubicBezSurf()
            self.Surfs[nid].SurfFromCntrlPts(pts)

        fr.Next()
        if fr[0] != "NUM_FRONTS:": raise "FormatException"
        num_fronts = fr.AsInt(1)

        for i in xrange(num_fronts):
            fr.Next()
            if fr[0] != "FRONT:": raise "FormatException"
            num_nds = fr.AsInt(2)
            tmp = []
            while(1):
                fr.Next()
                for k in xrange(len(fr)): 
                    tmp.append(fr.AsInt(k))
                if len(tmp) >= num_nds: break
            self.Fronts.append(tmp)

            fr.Next()
            if fr[0] != "RADII:": raise "FormatException"
            num_rad = fr.AsInt(2)
            tmpd = []
            while(1):
                fr.Next()
                for k in xrange(len(fr)):
                    tmpd.append(fr.AsFloat(k))
                if len(tmpd) >= num_rad: break
            self.FrontTempRadii.append(tmpd)

            fr.Next()
            if fr[0] != "TEMP_PAR:": raise "FormatException"
            if fr.AsInt(7) != 0:
                tmp0 = True
            else:
                tmp0 = False
            if version == 1:
                tmp1 = False
            elif version == 2:
                if fr.AsInt(8) != 0:
                    tmp1 = True
                else:
                    tmp1 = False
            self.FrontTempParams.append(
                self.TempParams(fr.AsFloat(2),fr.AsInt(3),fr.AsInt(4),
                                fr.AsFloat(5),fr.AsFloat(6),tmp0,tmp1))


        fr.Next()
        if fr[0] != ")": raise "FormatException"
        self._GenerateNodeCds()


    def RefineUniform(self):

        # The purpose of this method is to uniformly divide all
        # triangular facets into four smaller facets.  The local
        # vertex and edge numbering is as follows
        #
        #          0                          o
        #         / \                        / \
        #        /   \                      0   5
        #       /     \                    /     \
        #      3-------5                  o---8---o
        #     / \     / \                / \     / \
        #    /   \   /   \              1   6   7   4
        #   /     \ /     \            /     \ /     \
        #  1-------4-------2          o---2---o---3---o

        new_conn = ((0,3,5),(3,1,4),(4,2,5),(3,4,5))
        new_edge = ((0,8,5),(1,2,6),(3,4,7),(6,7,8))
        new_outer_edge = ((0,3),(3,1),(1,4),(4,2),(2,5),(5,0))
        new_inner_edge = (((0.333333333,0.166666667),(0.166666667,0.333333333)),
                          ((0.5,        0.333333333),(0.5,        0.166666667)),
                          ((0.166666667,0.5        ),(0.333333333,0.5        ))) ;

        pars = (Vec2D.Vec2D(0.0,1.0),Vec2D.Vec2D(0.0,0.0),Vec2D.Vec2D(1.0,0.0))

        # First store information about the edges and verts

        edges = {}
        verts = {}
        vert_coords = {}

        for i in self.Nodes:
            for j in xrange(3):

                if self.Nodes[i][j] >= self.NextNodeId:
                    self.NextNodeId = self.Nodes[i][j] + 1

                k = (j+1) % 3
                key = EdgeKey(self.Nodes[i][j],self.Nodes[i][k]) ;

                if edges.has_key(key):
                    edges[key].face1 = i
                    edges[key].face1_par[0] = pars[j]
                    edges[key].face1_par[1] = pars[k]
                else:
                    edges[key] = self.Edge(self.Nodes[i][j],self.Nodes[i][k],i,-1)
                    edges[key].face0_par[0] = pars[j]
                    edges[key].face0_par[1] = pars[k]

                if verts.has_key(self.Nodes[i][j]):
                    verts[self.Nodes[i][j]].append(self.VtxRef(i,pars[j]))
                else:
                    tmp = []
                    tmp.append(self.VtxRef(i,pars[j]))
                    verts[self.Nodes[i][j]] = tmp
                    vert_coords[self.Nodes[i][j]] = self.Surfs[i].Evaluate(pars[j])

        # compute the normals based on the average normal of the
        # adjacent faces.  We should probably update this to weight
        # the normal by the angle.  Also store the coordinates of the
        # corner nodes.

        vert_normals = {}
        for key in verts:
            normal = Vec3D.Vec3D(0,0,0)
            tmp = verts[key]
            for i in xrange(len(tmp)):
                normal += self.Surfs[tmp[i].face].Normal(tmp[i].param)
            vert_normals[key] = (normal/len(tmp)).Normalize()
            vert_coords[key] = self.Surfs[tmp[0].face].Evaluate(tmp[0].param)
 
        # now for all the edges compute the coordinate of the mid-side
        # vertex and store the coordinates of points along the edges that
        # will be used to define the new sub-face geometry

        edge_pts = {}
        for key in edges:
            entry = edges[key]
            entry.side_node = self.NextNodeId

            # compute the mid-side coordinate

            edge = entry
            mid = 0.5 * (edge.face0_par[0]+edge.face0_par[1])
            vert_coords[self.NextNodeId] = self.Surfs[edge.face0].Evaluate(mid)
            self.NodeCds[self.NextNodeId] = vert_coords[self.NextNodeId]

            # compute the mid-side normal

            normal = self.Surfs[edge.face0].Normal(mid)
            if edge.face1 >= 0:
                mid  = 0.5 * (edge.face1_par[0]+edge.face1_par[1])
                normal += self.Surfs[edge.face1].Normal(mid)
                normal /= 2

            vert_normals[self.NextNodeId] = normal.Normalize()
            self.NodeNrm[self.NextNodeId] = vert_normals[self.NextNodeId]

            # now the edge coordinate points

            mid0 = (5.0*edge.face0_par[0] + edge.face0_par[1]) / 6.0
            mid1 = (4.0*edge.face0_par[0] + 2.0*edge.face0_par[1]) / 6.0
            edge_pts[EdgeKey(edge.vt0,edge.side_node)] = \
                      self.EdgePts(self.Surfs[edge.face0].Evaluate(mid0),
                                   self.Surfs[edge.face0].Evaluate(mid1),
                                   edge.face0)

            mid0 = (2.0*edge.face0_par[0] + 4.0*edge.face0_par[1]) / 6.0
            mid1 = (edge.face0_par[0] + 5.0*edge.face0_par[1]) / 6.0
            edge_pts[EdgeKey(edge.side_node,edge.vt1)] = \
                      self.EdgePts(self.Surfs[edge.face0].Evaluate(mid0),
                                   self.Surfs[edge.face0].Evaluate(mid1),
                                   edge.face0)
            self.NextNodeId += 1

        # here we define the new faces

        NewNodes = {}
        NewSurfs = {}

        flags = [True,True,True]

        for i in self.Nodes:

            # get the 6 nodes for the subdivided facet

            node_id = [None]*6
            for j in xrange(3):
                k = (j+1) % 3
                node_id[j] = self.Nodes[i][j] ;
                node_id[j+3] = edges[EdgeKey(self.Nodes[i][j],self.Nodes[i][k])].side_node

            # get the points that define the edges for the new facets.  First
            # the edges that are defined by the "parent" geometry

            edge_coords = []
            for j in xrange(9):
                edge_coords.append([Vec3D.Vec3D(0,0,0),Vec3D.Vec3D(0,0,0)])
            for j in xrange(6):
                key = EdgeKey(node_id[new_outer_edge[j][0]],
                              node_id[new_outer_edge[j][1]])
                if edge_pts[key].face == i:
                    edge_coords[j][0] = edge_pts[key].pt0
                    edge_coords[j][1] = edge_pts[key].pt1
                else:
                    edge_coords[j][0] = edge_pts[key].pt1
                    edge_coords[j][1] = edge_pts[key].pt0

            # now the edges that cross the parent surface

            for j in xrange(3):
                edge_coords[j+6][0] = self.Surfs[i].Evaluate(new_inner_edge[j][0][0],
                                                             new_inner_edge[j][0][1])
                edge_coords[j+6][1] = self.Surfs[i].Evaluate(new_inner_edge[j][1][0],
                                                             new_inner_edge[j][1][1])

            # now define the new facets

            for j in xrange(4):

                vts = (node_id[new_conn[j][0]],
                       node_id[new_conn[j][1]],
                       node_id[new_conn[j][2]])

                if j == 0:
                    NewNodes[i] = vts
                else:
                    NewNodes[self.NextSurfId] = vts

                corners = [Vec3D.Vec3D(0,0,0)]*3
                normals = [Vec3D.Vec3D(0,0,0)]*3
                bdry_pts = []
                for k in xrange(3): bdry_pts.append([Vec3D.Vec3D(0,0,0),Vec3D.Vec3D(0,0,0)])

                for k in xrange(3):
                    corners[k] = vert_coords[vts[k]]
                    normals[k] = vert_normals[vts[k]]
                    if j < 3:
                        bdry_pts[k][0] = edge_coords[new_edge[j][k]][0]
                        bdry_pts[k][1] = edge_coords[new_edge[j][k]][1]
                    else :
                        bdry_pts[k][0] = edge_coords[new_edge[j][k]][1]
                        bdry_pts[k][1] = edge_coords[new_edge[j][k]][0]

                surf = CubicBezSurf.CubicBezSurf()
                #print >> sys.stderr,flags
                surf.SurfFromMixedBdry(corners,normals,bdry_pts,flags)

                if j == 0:
                    NewSurfs[i] = surf
                else:
                    NewSurfs[self.NextSurfId] = surf
                    self.NextSurfId += 1

        # do the crack fronts

        NewFronts = []
        NewTempRadii = []
        for i in xrange(len(self.Fronts)):
            tmp = []
            rad = []
            tmp.append(self.Fronts[i][0])
            rad.append(self.FrontTempRadii[i][0])
            last = self.FrontTempRadii[i][0]
            for j in xrange(len(self.Fronts[i])-1):
                edge = edges[EdgeKey(self.Fronts[i][j],self.Fronts[i][j+1])]
                tmp.append(edge.side_node)
                rad.append(0.5*(last+self.FrontTempRadii[i][j]))
                last = self.FrontTempRadii[i][j]
                if edge.vt0 == self.Fronts[i][j]:
                    tmp.append(edge.vt1)
                else:
                    tmp.append(edge.vt0)
                rad.append(last)
            NewFronts.append(tmp)
            NewTempRadii.append(rad)

        self.Nodes = NewNodes
        self.Surfs = NewSurfs
        self.Fronts = NewFronts
        self.FrontTempRadii = NewTempRadii


    def Rotate(self,axis,angle):
        angle_rad = angle * 0.0174532925199433
        ca = math.cos(angle_rad)
        sa = math.sin(angle_rad)
        rot = [Vec3D.Vec3D(1,0,0),Vec3D.Vec3D(0,1,0),Vec3D.Vec3D(0,0,1)]
        if axis == 0:
            rot[1][1] =  ca
            rot[1][2] = -sa
            rot[2][1] =  sa
            rot[2][2] =  ca
        elif axis == 1:
            rot[0][0] =  ca
            rot[0][2] =  sa
            rot[2][0] = -sa
            rot[2][2] =  ca
        elif axis == 2:
            rot[0][0] =  ca
            rot[0][1] = -sa
            rot[1][0] =  sa
            rot[1][1] =  ca

        for nid in self.NodeCds:
            coord = self.NodeCds[nid]
            norm = self.NodeNrm[nid]
            self.NodeCds[nid] = Vec3D.Vec3D(rot[0]*coord,rot[1]*coord,rot[2]*coord)
            self.NodeNrm[nid] = Vec3D.Vec3D(rot[0]*norm,rot[1]*norm,rot[2]*norm)

        for i in self.Surfs:
            for j in xrange(10):
                coord = self.Surfs[i].Pts[j]
                self.Surfs[i].Pts[j] = Vec3D.Vec3D(rot[0]*coord,rot[1]*coord,rot[2]*coord)


    def RotateArbAxis(self,axis,angle):
        angle_rad = angle * 0.0174532925199433
        c = math.cos(angle_rad)
        s = math.sin(angle_rad)
        t = 1 - math.cos(angle_rad)
        n = axis.Normalize()

        rot = [Vec3D.Vec3D(1,0,0),Vec3D.Vec3D(0,1,0),Vec3D.Vec3D(0,0,1)]
        rot[0][0] = t*n[0]**2 + c
        rot[0][1] = t*n[0]*n[1] + s*n[2]
        rot[0][2] = t*n[0]*n[2] - s*n[1]
        rot[1][0] = t*n[0]*n[1] - s*n[2]
        rot[1][1] = t*n[1]**2 + c
        rot[1][2] = t*n[1]*n[2] + s*n[0]
        rot[2][0] = t*n[0]*n[2] + s*n[1]
        rot[2][1] = t*n[1]*n[2] - s*n[0]
        rot[2][2] = t*n[2]**2 + c

        for nid in self.NodeCds:
            coord = self.NodeCds[nid]
            norm = self.NodeNrm[nid]
            self.NodeCds[nid] = Vec3D.Vec3D(rot[0]*coord,rot[1]*coord,rot[2]*coord)
            self.NodeNrm[nid] = Vec3D.Vec3D(rot[0]*norm,rot[1]*norm,rot[2]*norm)

        for i in self.Surfs:
            for j in xrange(10):
                coord = self.Surfs[i].Pts[j]
                self.Surfs[i].Pts[j] = Vec3D.Vec3D(rot[0]*coord,rot[1]*coord,rot[2]*coord)

    def Translate(self,trans):

        for nid in self.NodeCds:
            coord = self.NodeCds[nid]
            self.NodeCds[nid] = coord + trans

        for i in self.Surfs:
            for j in xrange(10):
                coord = self.Surfs[i].Pts[j]
                self.Surfs[i].Pts[j] = coord + trans


    def Plot(self,fd):
        self._GenerateNodeCds()
        self.MakeWatertight()
        for i in self.Surfs:
            self.Surfs[i].DebugDrawSurface(fd)
            self.Surfs[i].DebugDrawOutline(fd)
            cntr = self.Surfs[i].Evaluate(0.3333,0.3333)
            print >> fd, "t",cntr,i
            for pt in self.Surfs[i].Pts:
                print >> fd,"v",pt
        for nid in self.NodeCds:
            print >> fd, "v",self.NodeCds[nid]
            print >> fd, "t",self.NodeCds[nid],nid
        #for front in self.Fronts:
            #print >> sys.stderr,front
            #print >> sys.stderr,self.NodeCds
        #    for i in xrange(len(front)):
        #        print >> fd, "v", self.NodeCoord(front[i])

    def MakeWatertight(self):
        nm = ((1,3),(7,8),(5,2))
        edges = {}
        for i in self.Nodes:
            #if i == 38: print >> sys.stderr,"D",self.Nodes[i]
            for j in xrange(3):
                k = (j+1) % 3
                key = EdgeKey(self.Nodes[i][j],self.Nodes[i][k])
                if edges.has_key(key):
                    # average the control point coordinates
                    surf,side = edges[key]
                    for l in (0,1):
                        m = (l+1) % 2
                        #if i == 38: print >> sys.stderr,"D",nm[j][l],nm[side][m]
                        avg = 0.5*(self.Surfs[i].Pts[nm[j][l]] +
                                   self.Surfs[surf].Pts[nm[side][m]])
                        #if i == 38 and l == 0:
                        #    print "v",self.Surfs[i].Pts[nm[j][l]],"1 0 0"
                        #    print "v",self.Surfs[surf].Pts[nm[side][m]],"1 0 0"
                        self.Surfs[i].Pts[nm[j][l]] = avg
                        self.Surfs[surf].Pts[nm[side][m]] = avg
                else:
                    edges[key] = (i,j)

    def GetFrontPoints(self,num=20,front=-1):
        if front >= 0:
            return self._DoGetFrontPoints(num,front)
        else:
            npts = []
            cpts = []
            axes = []
            for i in xrange(len(Self.Fronts)):
                a,b,c = self._DoGetFrontPoints(num,i)
                npts.append(a)
                npts.append(b)
                npts.append(c)
            return npts,cpts,axes

    def GetEdgePoints(self,verts):

        # loop through the edges defined by the input vertices, find an
        # adjacent surface and edge, and evaluate some cartesian points

        out_pts = []
        surfs = []

        for i in xrange(len(verts)-1):
            pt0 = verts[i]
            pt1 = verts[i+1]

            sid = -1
            for sid in self.Nodes:
                snodes = self.Nodes[sid]
                for j in xrange(3):
                    if pt0 == snodes[j]:
                        if pt1 == snodes[(j+1)%3]:
                            surf_id = sid
                            surf_start = j
                            surf_direct = True
                            break
                        if pt1 == snodes[j-1]:
                            surf_id = sid
                            surf_start = j
                            surf_direct = False
                            break

            assert sid >= 0

            # evaluate the points

            surfs.append(self.Surfs[surf_id])
            edge_pars = [pt for pt in self.EdgeMap[surf_start]]
            if not surf_direct: edge_pars.Reverse()

            if len(out_pts) == 0:
                out_pts.append(self.Surfs[surf_id].Evaluate(edge_pars[0]))

            for j in xrange(1,4):
                out_pts.append(self.Surfs[surf_id].Evaluate(edge_pars[j]))

        return out_pts


    # private methods

    def _GenerateNodeCds(self):
        rs = [Vec2D.Vec2D(0.0,1.0),Vec2D.Vec2D(0.0,0.0),Vec2D.Vec2D(1.0,0.0)]
        for i in self.Nodes:
            for j in xrange(3):
                if not self.NodeCds.has_key(self.Nodes[i][j]):
                    self.NodeCds[self.Nodes[i][j]] = self.Surfs[i].Evaluate(rs[j])
                    self.NodeNrm[self.Nodes[i][j]] = self.Surfs[i].Normal(rs[j])

    def _DoGetFrontPoints(self,num,front):

        # loop through the edges on the front, find the surface, edge, and
        # evaluate some cartesian points we can use for constructing a spline

        spline_pts = []
        surfs = []

        for i in xrange(len(self.Fronts[front])-1):
            pt0 = self.Fronts[front][i]
            pt1 = self.Fronts[front][i+1]

            sid = -1
            for sid in self.Nodes:
                snodes = self.Nodes[sid]
                for j in xrange(3):
                    if pt0 == snodes[j]:
                        if pt1 == snodes[(j+1)%3]:
                            surf_id = sid
                            surf_start = j
                            surf_direct = True
                            break
                        if pt1 == snodes[j-1]:
                            surf_id = sid
                            surf_start = j
                            surf_direct = False
                            break

            assert sid >= 0

            # evaluate the points

            surfs.append(self.Surfs[surf_id])
            edge_pars = [pt for pt in self.EdgeMap[surf_start]]
            if not surf_direct: edge_pars.Reverse()

            if len(spline_pts) == 0:
                spline_pts.append(self.Surfs[surf_id].Evaluate(edge_pars[0]))

            for j in xrange(1,4):
                spline_pts.append(self.Surfs[surf_id].Evaluate(edge_pars[j]))

        # Generate the spline

        spline = BSpline3D.BSpline3D(spline_pts)

        # now evaluate the spline to find points, and axes

        npts = []
        cpts = []
        axes = []

        for i in xrange(num+1):
            par = float(i) / float(num)
            npts.append(par)
            cpts.append(spline.Evaluate(par))

            # find the closest surface point and normal

            first = True
            min_dist = None

            for surf in surfs:
                good,close,r,s,dist = surf.FindClosestPoint(cpts[-1])
                #print >> sys.stderr,good,dist,min_dist
                if not good: continue ;
                if first or dist < min_dist:
                    min_surf = surf
                    min_r = r
                    min_s = s
                    min_dist = dist
                    first = False

            assert first == False

            norm = min_surf.Normal(min_r,min_s).Normalize()
            #print "e",cpts[-1],cpts[-1]+(0.005*norm),"1 0 0"
            #min_surf.DebugDrawSurface(sys.stdout)
            #min_surf.DebugDrawOutline(sys.stdout)
            #raise SystemExit
            tang = spline.Tangent(par).Normalize()
            x = Vec3D.CrossProd(norm,tang).Normalize()
            z = Vec3D.CrossProd(x,norm).Normalize()
            axes.append((x,norm,z))

        return npts,cpts,axes

        
